#include "yuv.h"

/* how many significant bits to store for the color information for yuv */
static int _y_bits = 8;
static int _u_bits = 8;
static int _v_bits = 8;

/* shortcut speedup lookup-tables */
static unsigned char _v257[256];
static unsigned char _v504[256];
static unsigned char _v098[256];
static unsigned char _v439[256];
static unsigned char _v368[256];
static unsigned char _v071[256];
static unsigned char _v148[256];
static unsigned char _v291[256];
static unsigned char _w439[256];

static int _v1164[256];
static int _v1596[256];
static int _v813[256];
static int _v391[256];
static int _v2018[256];

static int _ay16  = 16;
static int _ay128 = 128;
static int _au16  = 16;
static int _au128 = 128;
static int _av16  = 16;
static int _av128 = 128;

static unsigned char _clip_lut[1024];
#define LUT_CLIP(i) ((_clip_lut+384)[(i)])

#define CMP_CLIP(i) ((i&256)? (~(i>>10)) : i);

/* functions */
void
yuv_init(int ybits, int ubits, int vbits)
{
  int i;
  int masky, masku, maskv;
  
  _y_bits = ybits;
  _u_bits = ubits;
  _v_bits = vbits;
  
  _ay16  = 16  >> (8 - _y_bits); 
  _ay128 = 128 >> (8 - _y_bits); 
  _au16  = 16  >> (8 - _u_bits); 
  _au128 = 128 >> (8 - _u_bits); 
  _av16  = 16  >> (8 - _v_bits); 
  _av128 = 128 >> (8 - _v_bits); 
  
  masky = 0;
  for (i = 0 ; i < _y_bits; i++)
    masky |= 1 << i;
  masku = 0;
  for (i = 0 ; i < _u_bits; i++)
    masku |= 1 << i;
  maskv = 0;
  for (i = 0 ; i < _v_bits; i++)
    maskv |= 1 << i;
  
  for (i = 0; i < 256; i++)
    {
      _v257[i] = (unsigned char)((float)i * 0.257) >> (8 - _y_bits);
      _v504[i] = (unsigned char)((float)i * 0.504) >> (8 - _y_bits);
      _v098[i] = (unsigned char)((float)i * 0.098) >> (8 - _y_bits);
      
      _v439[i] = (unsigned char)((float)i * 0.439) >> (8 - _u_bits);
      _v368[i] = (unsigned char)((float)i * 0.368) >> (8 - _u_bits);
      _v071[i] = (unsigned char)((float)i * 0.071) >> (8 - _u_bits);
      
      _v148[i] = (unsigned char)((float)i * 0.148) >> (8 - _v_bits);
      _v291[i] = (unsigned char)((float)i * 0.291) >> (8 - _v_bits);
      _w439[i] = (unsigned char)((float)i * 0.439) >> (8 - _v_bits);

      _v1164[i] = (int)(((float)(i - _ay16 )) * 1.164) << (8 - _y_bits);
      _v1164[i] |= (_v1164[i] >> _y_bits) & (masky >> _y_bits);
      
      _v1596[i] = (int)(((float)(i - _au128)) * 1.596) << (8 - _u_bits);
      _v1596[i] |= (_v1596[i] >> _u_bits) & (masku >> _u_bits);
      _v813[i]  = (int)(((float)(i - _au128)) * 0.813) << (8 - _u_bits);
      _v813[i] |= (_v813[i] >> _u_bits) & (masku >> _u_bits);
      
      _v391[i]  = (int)(((float)(i - _av128)) * 0.391) << (8 - _v_bits);
      _v391[i] |= (_v391[i] >> _v_bits) & (maskv >> _v_bits);
      _v2018[i] = (int)(((float)(i - _av128)) * 2.018) << (8 - _v_bits);
      _v2018[i] |= (_v2018[i] >> _v_bits) & (maskv >> _v_bits);
    }

   for (i = -384; i < 640; i++) 
     _clip_lut[i+384] = i < 0 ? 0 : (i > 255) ? 255 : i;
}

void
yv12torgb_diz(unsigned char *yuv, unsigned char *rgb, int w, int h)
{
   int xx, yy;
   int y, u, v, r, g, b;   
   unsigned char *yp1, *yp2, *up, *vp;
   unsigned char *dp1, *dp2;
   int w2, h2;
   int crv, cbu, cgu, cgv;
   
   /* handy halved w & h */
   w2 = w / 2;
   h2 = h / 2;
   /* plane pointers */
   yp1 = yuv;
   yp2 = yuv + w;
   up = yuv + (w * h);
   vp = up + (w2 * h2);
   /* destination pointers */
   dp1 = rgb;
   dp2 = rgb + (w * 4);
   
   crv = 104595;   /* 1.596 */
   cbu = 132251;   /* 2.018 */
   cgu = 25624;    /* 0.391 */
   cgv = 53280;    /* 0.813 */
   
   for (yy = 0; yy < h2; yy++)
     {
	for (xx = 0; xx < w2; xx++)
	  {
	     /* collect u & v for 2x2 pixel block */
	     u = (*up++) - 128;
	     v = (*vp++) - 128;
	     
	     /* do the top 2 pixels of the 2x2 block whcih shared u & v */
	     /* yuv to rgb */
	     y = 76283 * ((*yp1++) - 16);
	     r = (y + (crv * v)) >> 16;
	     r = LUT_CLIP(r);
	     R_VAL(dp1) = r;
	     g = (y - (cgu * u) - (cgv * v) + 32768) >>16;
	     g = LUT_CLIP(g);
	     G_VAL(dp1) = g;
	     b = (y + (cbu * u) + 32768) >> 16;
	     b = LUT_CLIP(b);
	     B_VAL(dp1) = b;
	     
	     dp1 += 4;
	     
	     /* yuv to rgb */
	     y = 76283 * ((*yp1++) - 16);
	     r = (y + (crv * v)) >> 16;
	     r = LUT_CLIP(r);
	     R_VAL(dp1) = r;
	     g = (y - (cgu * u) - (cgv * v) + 32768) >>16;
	     g = LUT_CLIP(g);
	     G_VAL(dp1) = g;
	     b = (y + (cbu * u) + 32768) >> 16;
	     b = LUT_CLIP(b);
	     B_VAL(dp1) = b;
	     
	     dp1 += 4;
	     
	     /* do the bottom 2 pixels */
	     /* yuv to rgb */
	     y = 76283 * ((*yp2++) - 16);
	     r = (y + (crv * v)) >> 16;
	     r = LUT_CLIP(r);
	     R_VAL(dp2) = r;
	     g = (y - (cgu * u) - (cgv * v) + 32768) >>16;
	     g = LUT_CLIP(g);
	     G_VAL(dp2) = g;
	     b = (y + (cbu * u) + 32768) >> 16;
	     b = LUT_CLIP(b);
	     B_VAL(dp2) = b;

	     dp2 += 4;
	     
	     /* yuv to rgb */
	     y = 76283 * ((*yp2++) - 16);
	     r = (y + (crv * v)) >> 16;
	     r = LUT_CLIP(r);
	     R_VAL(dp2) = r;
	     g = (y - (cgu * u) - (cgv * v) + 32768) >>16;
	     g = LUT_CLIP(g);
	     G_VAL(dp2) = g;
	     b = (y + (cbu * u) + 32768) >> 16;
	     b = LUT_CLIP(b);
	     B_VAL(dp2) = b;

	     dp2 += 4;
	  }
	/* jump down one line since we are doing 2 at once */
	yp1 += w;
	yp2 += w;
	dp1 += (w * 4);
	dp2 += (w * 4);
     }
}

void
yv12torgb_raster(unsigned char *yuv, unsigned char *rgb, int w, int h)
{
   int xx, yy;
   int y, u, v, r, g, b;
   unsigned char *yp1, *yp2, *up, *vp;
   unsigned char *dp1, *dp2;
   int w2, h2;
   
   /* handy halved w & h */
   w2 = w / 2;
   h2 = h / 2;
   /* plane pointers */
   yp1 = yuv;
   yp2 = yuv + w;
   up = yuv + (w * h);
   vp = up + (w2 * h2);
   /* destination pointers */
   dp1 = rgb;
   dp2 = rgb + (w * 4);
   for (yy = 0; yy < h2; yy++)
     {
	for (xx = 0; xx < w2; xx++)
	  {
	     /* collect u & v for 2x2 pixel block */
	     u = *up++;
	     v = *vp++;
	     
	     /* do the top 2 pixels of the 2x2 block whcih shared u & v */
	     /* yuv to rgb */
	     y = *yp1++;
	     r = _v1164[y] + _v1596[v];
	     r = CMP_CLIP(r);
	     g = _v1164[y] -  _v813[v] -  _v391[u];
	     g = CMP_CLIP(g);
	     b = _v1164[y] + _v2018[u];
	     b = CMP_CLIP(b);
	     
	     R_VAL(dp1) = r;
	     G_VAL(dp1) = g;
	     B_VAL(dp1) = b;
	     
	     dp1 += 4;
	     
	     /* yuv to rgb */
	     y = *yp1++;
	     r = _v1164[y] + _v1596[v];
	     r = CMP_CLIP(r);
	     g = _v1164[y] -  _v813[v] -  _v391[u];
	     g = CMP_CLIP(g);
	     b = _v1164[y] + _v2018[u];
	     b = CMP_CLIP(b);
	     
	     R_VAL(dp1) = r;
	     G_VAL(dp1) = g;
	     B_VAL(dp1) = b;
	     
	     dp1 += 4;
	     
	     /* do the bottom 2 pixels */
	     /* yuv to rgb */
	     y = *yp2++;
	     r = _v1164[y] + _v1596[v];
	     r = CMP_CLIP(r);
	     g = _v1164[y] -  _v813[v] -  _v391[u];
	     g = CMP_CLIP(g);
	     b = _v1164[y] + _v2018[u];
	     b = CMP_CLIP(b);
	     
	     R_VAL(dp2) = r;
	     G_VAL(dp2) = g;
	     B_VAL(dp2) = b;

	     dp2 += 4;
	     
	     /* yuv to rgb */
	     y = *yp2++;
	     
	     r = _v1164[y] + _v1596[v];
	     r = CMP_CLIP(r);
	     g = _v1164[y] -  _v813[v] -  _v391[u];
	     g = CMP_CLIP(g);
	     b = _v1164[y] + _v2018[u];
	     b = CMP_CLIP(b);
	     
	     R_VAL(dp2) = r;
	     G_VAL(dp2) = g;
	     B_VAL(dp2) = b;
	     
	     dp2 += 4;
	  }
	/* jump down one line since we are doing 2 at once */
	yp1 += w;
	yp2 += w;
	dp1 += (w * 4);
	dp2 += (w * 4);
     }
}
